/*
   Copyright 2006-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.notuvy.net;

import java.io.*;
import java.awt.*;

import org.apache.commons.httpclient.*;
import org.apache.commons.httpclient.auth.AuthScope;
import org.apache.commons.httpclient.methods.GetMethod;

import org.apache.log4j.Logger;

import com.notuvy.cmd.Command;
import com.notuvy.cmd.CommandArgs;
import com.notuvy.cmd.CommandAction;
import com.notuvy.util.Progressable;
import com.notuvy.util.DummyProgress;
import com.notuvy.thread.ThreadMonitor;

/**
 * A very simple utility for downloading a resource from an URL.
 *
 * @author  murali
 */
public class Download implements Runnable {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger(Download.class);

    private static final int BUFFER_SIZE = 1024;
    private static final int SOCKET_READ_TIMEOUT = 60000;  // one minute.


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private final String vUrl;
    private final File vFile;
    private Progressable vProgress = new DummyProgress();
    private boolean vIsAborted = false;
    private boolean vIsSuccessful = false;
    private int fStatusCode = -1;
    private UsernamePasswordCredentials fBasicCredentials = null;


    //------------------------------------------------------------
    //- Constructors

    public Download(String pUrl) {
        this (pUrl, null);
    }

    public Download(String pUrl, File pFile) {
        vUrl = pUrl;
        vFile = pFile;
    }


    //------------------------------------------------------------
    //- Accessors

    public String getUrl() {
        return vUrl;
    }

    public File getFile() {
        return vFile;
    }

    public Progressable getProgress() {
        return vProgress;
    }

    public boolean getIsAborted() {
        return vIsAborted;
    }

    public boolean getIsSuccessful() {
        return vIsSuccessful;
    }

    public void setStatusCode(int pStatusCode) {
        fStatusCode = pStatusCode;
    }


    //------------------------------------------------------------
    //- Settors

    public void setProgress(Progressable pProgress) {
        vProgress = pProgress;
    }

    public void setIsAborted(boolean pIsAborted) {
        vIsAborted = pIsAborted;
    }

    public void setIsSuccessful(boolean pIsSuccessful) {
        vIsSuccessful = pIsSuccessful;
    }

    public int getStatusCode() {
        return fStatusCode;
    }


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    private void toOutputStream(HttpMethod pMethod, OutputStream pOutput) throws IOException {
        try {
            int contentLength = 0;
            Header header = pMethod.getResponseHeader("Content-Length");
            if (header != null) {
                contentLength = Integer.parseInt(header.getValue());
                getProgress().setMaximum(contentLength);
            }

            byte[] buffer = new byte[BUFFER_SIZE];

            InputStream stream = pMethod.getResponseBodyAsStream();

            int bytesRead=stream.read(buffer);
            while (!getIsAborted() && (bytesRead != -1)) {
                getProgress().incrementValue(bytesRead);
                pOutput.write(buffer, 0, bytesRead);
                bytesRead=stream.read(buffer);
            }

            pOutput.close();

            boolean partialRead = (!getIsAborted() && (contentLength > 0) && (getProgress().getValue() < contentLength));
            if (partialRead) {
                LOG.error(String.format("Expected [%d] bytes but got only [%d].", contentLength, getProgress().getValue()));
                // Need to set the aborted flag so the file gets deleted.
                setIsAborted(true);
            }

            if (!getIsAborted()) {
                setIsSuccessful(true);
            }
        } finally {
            getProgress().finished();
        }
    }


    //------------------------------------------------------------
    //- Public Interface Functions

    public void setBasicCredentials(String pUser, String pPass) {
        fBasicCredentials = new UsernamePasswordCredentials(pUser, pPass);
    }

    public String giveStringBody() {
        final ByteArrayOutputStream output = new ByteArrayOutputStream();

        DownloadHandler handler = new DownloadHandler(new GetMethod(getUrl())) {
            public void handle() throws IOException {
                toOutputStream(getMethod(), output);
            }
        };
        handler.execute();

        return getIsSuccessful() ? output.toString() : null;
    }


    public void run() {
        ThreadMonitor.update("Fired.");

        DownloadHandler handler = new DownloadHandler(new GetMethod(getUrl())) {
            @Override
            protected void extraFinally() {
                ThreadMonitor.update("Finished.");
            }

            public void handle() throws IOException {
                FileOutputStream output = new FileOutputStream(getFile());
                toOutputStream(getMethod(), output);

                if (getIsAborted()) {
                    if (getFile().exists() && !getFile().delete()) {
                        LOG.warn(String.format("Problem deleting file [%s].", getFile().getAbsolutePath()));
                    }
                    LOG.error(String.format("Download of [%s] aborted.", getFile()));
                }
            }
        };
        handler.execute();
    }


    public void runGui(String pTitle) {
        if (GraphicsEnvironment.isHeadless()) {
            LOG.warn("GraphicsEnvironment is headless.  Running in non-GUI mode.");
        } else {
            DownloadGui gui = new DownloadGui(pTitle, this);
            gui.launch();
        }
        run();
    }


    //------------------------------------------------------------
    //- Class Interface Functions



    //------------------------------------------------------------
    //- Inner Classes

    private abstract class DownloadHandler {
        private final HttpMethod fMethod;

        protected DownloadHandler(HttpMethod pMethod) {
            fMethod = pMethod;
        }

        public HttpMethod getMethod() {
            return fMethod;
        }

        public abstract void handle() throws IOException;

        protected void extraFinally() {
            // Subclasses can override to add more actions.
        }

        private void execute() {
            HttpClient client = new HttpClient();
            client.getParams().setSoTimeout(SOCKET_READ_TIMEOUT);

            if (fBasicCredentials != null) {
                client.getParams().setAuthenticationPreemptive(true);
                client.getState().setCredentials(AuthScope.ANY, fBasicCredentials);
            }

            try {
                setStatusCode(client.executeMethod(getMethod()));
                if (getStatusCode() == HttpStatus.SC_OK) {
                    handle();
                } else {
                    LOG.error(String.format("Problem downloading[%s]; response[%d].", getUrl(), getStatusCode()));
                }
            } catch (IOException ioe) {
                LOG.error("HTTP downloading problem.", ioe);
            } finally {
                getMethod().releaseConnection();
                getProgress().finished();
                extraFinally();
            }
        }
    }



    //------------------------------------------------------------
    //- Main

    public static void main(String[] args) {
        try {
            CommandAction commands = new Command("", "<url> <filename> [username/password]") {
                public boolean process(CommandArgs pArgs) {
                    boolean success = false;  // Pessimistic.
                    if (pArgs.hasNext()) {
                        String url = pArgs.next();
                        if (pArgs.hasNext()) {
                            String filename = pArgs.next();
                            Download download = new Download(url, new File(filename));

                            boolean credentialsOk = true;
                            if (pArgs.hasNext()) {
                                String credentials = pArgs.next();
                                String[] parts = credentials.split("/");
                                if (parts.length == 2) {
                                    download.setBasicCredentials(parts[0], parts[1]);
                                } else {
                                    credentialsOk = false;
                                }
                            }

                            if (credentialsOk && pArgs.exhausted()) {
                                download.runGui("Downloading");
                                success = true;
                            }
                        } else {
                            LOG.error("No filename given.");
                        }
                    } else {
                        LOG.error("No url given.");
                    }
                    return success;
                }
            };
            commands.process(args);
        } catch (Throwable thr) {
            LOG.error("Fatal error.", thr);
        }
    }


}
